# Spica::File::Image
# == Author(s) : Takuya Otani <takuya.otani@gmail.com> ==
# == Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. ==
# == Copyright (C) 2005 Spica Project Kei/Fuco/Takkyun ==

# ======================================
package Spica::File::Image;
use strict;
use Spica::File::Object;
use Spica::File;
use base qw( Spica::File );
use vars qw( $VERSION );
$VERSION = '0.03';
# 0.03 [2006/07/04] got rid of "return( undef )"
# 0.02 [2006/03/17] changed bracket rule
# 0.01 [2006/01/13] added rotate_image
# 0.00 [2005/10/27] generated
# --------------------------------------
# // defines / static variables
my %aSupportedImage = (
	'.jpg' => \&_get_size_jpg,
	'.png' => \&_get_size_png,
	'.gif' => \&_get_size_gif,
	'.bmp' => \&_get_size_bmp,
	'.tif' => \&_get_size_tif,
	'.swf' => \&_get_size_swf,
);
# --------------------------------------
# // public functions
sub is_image
{
	my $self = shift;
	return ( $self->content_type =~ /^image/ );
}
sub is_resizable
{
	my $self = shift;
	my $ext  = $self->supported_extension;
	return ( defined($aSupportedImage{$ext}) );
}
sub get_size
{
	my $self = shift;
	my %param = (
		'max_w' => 0, # maximum width
		'max_h' => 0, # maximum height
		@_
	);
	my ($w,$h) = (0,0);
	if ($self->is_resizable)
	{
		my $ext = $self->supported_extension;
		($w,$h) = &{$aSupportedImage{$ext}}($self->fullpath) if ($ext);
		($w,$h) = &_resize($w,$h,$param{'max_w'},$param{'max_h'}) if ($param{'max_w'} and $param{'max_h'});
	}
	return($w,$h);
}
sub create_smallimage
{ # create a small image via Image::Magick
	my $self = shift;
	my %param = (
		'path'       => $self->path,
		'name'       => undef,
		'size'       => [undef,undef], # width, height
		'permission' => undef,
		@_
	);
	return if (!$self->is_image);
	# return if ($param{'name'} eq '' or $param{'name'} eq $self->name); # [note] not necessary
	return if ($param{'size'}->[0] <= 0 or $param{'size'}->[1] <= 0);
	my $new_file = Spica::File::Object->new(
		'path' => $param{'path'},
		'name' => $param{'name'},
	);
	return if ($new_file->exists);
	eval
	{
		require Image::Magick;
		my $current = Image::Magick->new;
		$current->Read($self->fullpath);
		my $small = $current->Clone();
		$small->Scale('width'=>$param{'size'}->[0],'height'=>$param{'size'}->[1]);
		$small->Write($new_file->fullpath);
		$new_file->set_permission($param{'permission'});
	};
	return ($@) ? undef : $new_file;
}
sub rotate_image
{
	my $self = shift;
	my $angle = shift;
	eval
	{
		require Image::Magick;
		my $img = Image::Magick->new;
		$img->Read($self->fullpath);
		$img->Rotate('degrees'=>$angle);
		$img->Write($self->fullpath);
	};
	return ($@) ? undef : 1;
}
# --------------------------------------
# // private functions
sub _resize
{ # calculate size to remain the same aspect
	my ($w,$h,$max_w,$max_h) = @_;
	if ($w > $max_w or $h > $max_h)
	{
		my $tmp = (($max_w / $w) < ($max_h / $h)) ? $max_w / $w : $max_h / $h;
		$w = int($w * $tmp) or 1;
		$h = int($h * $tmp) or 1;
	}
	return($w,$h);
}
sub _get_size_jpg
{ # from KENT WEB <http://www.kent-web.com//>
	my $jpeg = shift;
	my ($t, $m, $c, $l, $w, $h);
	open(JPEG, "$jpeg") or return (0,0);
	binmode JPEG;
	read(JPEG, $t, 2);
	while (1)
	{
		read(JPEG, $t, 4);
		($m, $c, $l) = unpack("a a n", $t);
		if ($m ne "\xFF")
		{
			$w = $h = 0;
			last;
		}
		elsif ((ord($c) >= 0xC0) && (ord($c) <= 0xC3))
		{
			read(JPEG, $t, 5);
			($h, $w) = unpack("xnn", $t);
			last;
		}
		else
		{
			read(JPEG, $t, ($l - 2));
		}
	}
	close(JPEG);
	return($w, $h);
}
sub _get_size_gif
{ # from KENT WEB <http://www.kent-web.com//>
	my $gif = shift;
	my ($data);
	open(GIF,"$gif") or return (0,0);
	binmode(GIF);
	read(GIF,$data,10);
	close(GIF);
	$data = substr($data,-4) if ($data =~ /^GIF/);
	my $w = unpack("v",substr($data,0,2));
	my $h = unpack("v",substr($data,2,2));
	return($w, $h);
}
sub _get_size_bmp
{ # from Image::Size / Copyright (C) 2000 Randy J. Ray
	# <http://search.cpan.org/~rjray/Image-Size-2.992/Size.pm>
	my $bmp = shift;
	my $buffer = undef;
	open(BMP,"$bmp") or return(0, 0);
	binmode(BMP);
	read(BMP, $buffer, 26);
	my ($w, $h) = unpack("x18VV", $buffer);
	close(BMP);
	return($w, $h);
}
sub _get_size_tif
{ # from Image::Size / Copyright (C) 2000 Randy J. Ray
	# <http://search.cpan.org/~rjray/Image-Size-2.992/Size.pm>
	my $tif = shift;
	my $endian = 'n'; # default to big-endian
	my $header = undef;
	my $offset = undef;
	my $ifd = undef;
	# === preparation ===
	open(TIF,"$tif") or return(0, 0);
	binmode(TIF);
	read(TIF, $header, 4);
	$endian = 'v' if ($header =~ /II\x2a\x00/o); # little-endian
	my @packspec = (
		undef,       # nothing (shouldn't happen)
		'C',         # BYTE (8-bit unsigned integer)
		undef,       # ASCII
		$endian,     # SHORT (16-bit unsigned integer)
		uc($endian), # LONG (32-bit unsigned integer)
		undef,       # RATIONAL
		'c',         # SBYTE (8-bit signed integer)
		undef,       # UNDEFINED
		$endian,     # SSHORT (16-bit unsigned integer)
		uc($endian), # SLONG (32-bit unsigned integer)
	);
	read(TIF, $offset, 4);
	$offset = unpack(uc($endian), $offset);
	seek(TIF, $offset, 0); # set offset
	read(TIF, $ifd, 2); # get number of directory entries
	my $num_dirent = unpack($endian, $ifd); # Make it useful
	$offset += 2;
	$num_dirent = $offset + ($num_dirent * 12); # calc. maximum offset of IFD
	seek(TIF, $offset, 0); # set offset again
	# === search width and height information
	$ifd = '';
	my $tag = 0;
	my $type = 0;
	my $w = undef;
	my $h = undef;
	while ( !defined($w) or !defined($h) )
	{
		read(TIF, $ifd, 12);
		last if ($ifd eq '' or $offset > $num_dirent);
		$offset += 12;
		$tag = unpack($endian, $ifd);
		$type = unpack($endian, substr($ifd, 2, 2));
		next if ($type > @packspec + 0 or !defined($packspec[$type]));
		if ($tag == 0x0100)
		{ # width
			$w = unpack($packspec[$type], substr($ifd, 8, 4));
		}
		elsif ($tag == 0x0101)
		{ # height
			$h = unpack($packspec[$type], substr($ifd, 8, 4));
		}
	}
	close(TIF);
	return($w, $h);
}
sub _get_size_png
{ # from WWWis <http://www.bloodyeck.com/wwwis/>
	my $png = shift;
	my ($head, $a, $b, $c, $d, $e, $f, $g, $h);
	open(PNG,"$png") or return (0,0);
	binmode(PNG);
	if ( read(PNG, $head, 8) == 8 
	  && $head eq "\x89\x50\x4e\x47\x0d\x0a\x1a\x0a" 
	  && read(PNG, $head, 4) == 4 
	  && read(PNG, $head, 4) == 4 
	  && $head eq "IHDR" 
	  && read(PNG, $head, 8) == 8)
	  {
		($a,$b,$c,$d,$e,$f,$g,$h) = unpack("C"x8, $head);
	}
	else
	{
		return(0,0);
	}
	close(PNG);
	return($a<<24|$b<<16|$c<<8|$d, $e<<24|$f<<16|$g<<8|$h);
}
sub _get_size_swf
{ # from Image::Size / Copyright (C) 2000 Randy J. Ray
	# <http://search.cpan.org/~rjray/Image-Size-2.992/Size.pm>
	my $swf = shift;
	my $bin2int = sub { unpack("N", pack("B32", substr("0" x 32 . $_[0], -32))) };
	my $header = undef;
	open(SWF,"$swf") or return(0,0);
	binmode(SWF);
	read(SWF, $header, 3);
	seek(SWF, 0, 0);
	my ($w, $h, $bits, $bs, $buffer) = (0, 0, undef, undef, undef);
	if ($header =~ /^FWS/)
	{ # [non-compressed flash] adapted from code sent by Dmitry Dorofeev <dima@yasp.com>
		read(SWF, $buffer, 33);
		$bs = unpack('B133', substr($buffer, 8, 17)) if ($buffer ne '');
	}
	elsif ($header =~ /^CWS/)
	{ # [compressed flash] adapted from code sent by Victor Kuriashkin <victor@yasp.com>
		read(SWF, $buffer, 1058);
		if ($buffer ne '')
		{
			$buffer = substr($buffer, 8, 1024);
			$buffer = &_uncompress($buffer);
			$bs = unpack('B133', substr($buffer, 0, 9)) if ($buffer ne '');
		}
	}
	if ($bs ne '')
	{
		$bits = &$bin2int(substr($bs, 0, 5));
		$w = int( &$bin2int(substr($bs, 5 + $bits, $bits)) / 20 );
		$h = int( &$bin2int(substr($bs, 5 + $bits * 3, $bits)) / 20 );
	}
	close(SWF);
	return($w, $h);
}
sub _uncompress
{ # from Compress::Zlib / Copyright (c) 1995-2004 Paul Marquess
	my $in = shift;
	eval { require Compress::Zlib };
	return if ($@);
	my ($x,$err) = &Compress::Zlib::inflateInit();
	return ( $x and $err == Compress::Zlib::Z_OK() ) 
		? ($x->__unc_inflate($in))[0] 
		: undef;
}
1;
__END__
