# Spica::Session
# == Author(s) : Takuya Otani <takuya.otani@gmail.com> ==
# == Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. ==
# == Copyright (C) 2005 Spica Project Kei/Fuco/Takkyun ==

# ======================================
package Spica::Session;
use strict;
use Spica;
use Spica::Interface;
use Spica::Data;
use Spica::Object;
use base qw( Spica::Object );
use vars qw( $VERSION );
$VERSION = '0.02';
# 0.02 [2006/07/04] got rid of "return( undef )"
# 0.01 [2006/03/17] changed bracket rule
# 0.00 [2006/03/04] refactored
# --------------------------------------
# // defines / static variables
sub DEFAULT_EXPIRES (){ 1 };
sub COOKIENAME      (){ 'spica_session' };
sub HOURS_PER_DAY   (){ 24 };
sub SECS_PER_HOUR   (){ 3600 };
# --------------------------------------
# // override base class
sub initialize
{
	my $self = shift;
	my %param = (
		'key'    => undef,           # [required][CHAR] key for session
		'cgi'    => undef,           # [optional][OBJ.] Spica::Interface object
		'time'   => undef,           # [optional][NUM.] current time (UnixTime on GMT)
		'expire' => DEFAULT_EXPIRES, # [optional][NUM.] days for expires
		'name'   => COOKIENAME,      # [optional][CHAR] name for cookie
		'path'   => undef,           # [optional][URI.] path for cookie
		'domain' => undef,           # [optional][URI.] domain for cookie
		@_
	);
	$self->SUPER::initialize(%param);
	$self->{'cgi'} = Spica::Interface->get if (!$self->{'cgi'});
	$self->{'id'} = undef;
	$self->{'stored'} = undef;
	return $self->_initialize;
}
# --------------------------------------
# // public functions
sub id
{ # create session unique id
	my $self = shift;
	if ( !defined($self->{'id'}) )
	{
		$self->{'id'} = $self->{'time'} . '_' . Spica->get->pid;
	}
	return $self->{'id'};
}
sub check
{ # check session
	my $self = shift;
	return( defined($self->{'stored'}) );
}
sub start
{ # start session
	my $self = shift;
	$self->_set_data($self->id);
	$self->{'cgi'}->set_cookie(
		'time'   => $self->{'time'},
		'name'   => $self->{'name'},
		'expire' => $self->{'expire'},
		'path'   => $self->{'path'},
		'domain' => $self->{'domain'},
		'data'   => {'id' => $self->id},
	);
	return();
}
sub finish
{ # finish session
	my $self = shift;
	$self->_set_data(0);
}
# --------------------------------------
# // private functions
sub _initialize
{
	my $self = shift;
	if ( defined($self->{'key'}) )
	{
		$self->{'time'} ||= time(); # set the current time
		my $check_id = $self->_check_cookie;
		my $update_flag = undef;
		my @sessions = Spica::Data->load('Session','cond'=>{'key'=>$self->{'key'}});
		foreach my $session ( @sessions )
		{
			if ($check_id and $session->data eq $check_id)
			{
				$self->{'stored'} = $session;
			}
			if (!$session->check_expire('now'=>$self->{'time'},'duration'=>$self->_expire_time))
			{
				$session->erase;
				$update_flag = 1;
			}
		}
		Spica::Data->update(@sessions) if ($update_flag);
	}
	else
	{
		die("Spica::Session : Need to set key\n");
	}
	return $self;
}
sub _expire_time
{ # calculate expire time [sec]
	my $self = shift;
	return( $self->{'expire'} * HOURS_PER_DAY * SECS_PER_HOUR );
}
sub _check_cookie
{ # read cookie data
	my $self = shift;
	my $cookie = $self->{'cgi'}->cookie('name'=>$self->{'name'});
	if ( $cookie->{'id'} )
	{
		my $diff_time = $self->{'time'} - ( split('_',$cookie->{'id'}) )[0];
		return if ( $diff_time > $self->_expire_time );
	}
	return( $cookie->{'id'} );
}
sub _set_data
{ # update Spica::Data::Session
	my $self = shift;
	my $data = shift;
	my $session = ( $self->{'stored'} ) 
		? $self->{'stored'} 
		: Spica::Data->add('Session','key'=>$self->{'key'});
	die('failed to create session') if (!$session);
	$session->data($data); # set data
	Spica::Data->update($session);
}
1;
__END__
=head1 NAME

Spica::Session - session management

=head1 SYNOPSIS

	use Spica::Session;
	my $session = Spica::Session->new('key' => 'session_key');
	if ( $session->check )
	{
		# session is valid
		# something to do ...
		$session->finish; # finishing session
	}
	else
	{
		# session is invalid
		# something to do, such as user authorization
		$session->start; # starting session
	}

=head1 DESCRIPTION

Spica::Session requires Spica::Data::Session.
Session data is stored as Spica::Data::Object.
Need to set 'key' to create session.
'key' indicates each session, so Spica::Session allows multiple sessions.

=head1 AUTHOR

Takuya Otani <takuya.otani@gmail.com> 

=head1 COPYRIGHT & LICENSE

Copyright (C) 2005 Takuya Otani@SimpleBoxes / SerendipityNZ Ltd.

Copyright (C) 2005-2006 Kei / Fuco / Takkyun I<http://serennz.sakura.ne.jp/toybox/spica/>

Spica is distributed under the Perl Artistic Licence.
I<http://www.perl.com/pub/a/language/misc/Artistic.html>

=cut
