# Spica::Text
# == Author(s) : Takuya Otani <takuya.otani@gmail.com> ==
# == Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. ==
# == Copyright (C) 2005 Spica Project Kei/Fuco/Takkyun ==

# ======================================
package Spica::Text;
use strict;
use Spica::Object;
use Spica::Language;
use base qw( Spica::Object );
use vars qw( $VERSION $CAN_DIFF );
$VERSION = '0.12';
# 0.12 [2006/03/24] changed clip and _clip_text_utf8 to clip string correctly
# 0.11 [2006/03/17] changed bracket rule
# 0.10 [2006/03/09] changed _remove_tag_precise to fix a bug
# 0.09 [2006/01/10] changed uri_encoded/uri_decoded to escape slash(/).
# 0.08 [2005/12/31] changed diff to exchange source and destination
# 0.07 [2005/11/25] changed format/tag_removed to fix bugs
# 0.06 [2005/11/20] added uri_decoded/uri_encoded/linefeed
# 0.05 [2005/11/14] changed format to work as factory
# 0.04 [2005/11/06] added diff
# 0.03 [2005/10/21] refactored
# 0.02 [2005/07/07] added remove_tag as public method / renamed _remove_tag to _remove_tag_fast
# 0.01 [2005/06/06] changed entitize and detitize
# 0.00 [2005/02/17] generated
# --------------------------------------
# // initializer
BEGIN
{
	eval{ require Algorithm::Diff };
	$CAN_DIFF = ($@) ? undef : 1;
}
# --------------------------------------
# // defines / static variables
sub DEFAULT_CODE    (){ 'euc' };
sub CLIPPING_TAIL   (){ '...' };
sub CLIPPING_LENGTH (){ 200 };
sub SAME_LINE       (){ '=' };
sub ADDED_LINE      (){ '+' };
sub DELETED_LINE    (){ '-' };
# --------------------------------------
# // override base class
sub new
{
	my $class = shift;
	my $text = shift;
	return bless \$text, $class;
}
# --------------------------------------
# // public functions
sub data
{ # return raw data
	my $self = shift;
	$$self;
}
sub format
{ # format text
	my $self = shift;
	my $type = shift;
	my $option = shift;
	return( $$self ) if (!$type);
	my $text = $$self;
	eval
	{
		my $formatter = Spica::Text->module_name($type);
		$text = $formatter->$type($$self,$option) if (Spica::Text->load_module($type));
	};
	$$self = $text if (!$@);
	$$self;
}
sub clipped
{ # clip text
	my $self = shift;
	my %param = (
		'length'   => CLIPPING_LENGTH, # length for clipping
		'format'   => undef,           # format type
		'option'   => undef,           # format option
		'reversed' => undef,           # clipping position
		'ellipsis' => CLIPPING_TAIL,   # replaced string
		@_
	);
	if ( defined($param{'format'}) )
	{
		$self->format($param{'format'},$param{'option'});
		$$self = &_remove_tag_fast($$self);
	}
	$$self =~ tr/\x0D\x0A//d; # get rid of linefeeds
	if ( length($$self) > $param{'length'} )
	{
		$param{'length'} -= length($param{'ellipsis'});
		$$self = ($param{'reversed'})
			? $param{'ellipsis'} . &_clip_text($$self,-$param{'length'},$param{'length'})
			: &_clip_text($$self,0,$param{'length'}) . $param{'ellipsis'};
	}
	$$self;
}
sub entitized
{
	my $self = shift;
	$$self =~ s/&/&amp;/g;
	$$self =~ s/\"/&quot;/g;
	$$self =~ s/</&lt;/g;
	$$self =~ s/>/&gt;/g;
	$$self =~ s/\{/&#123;/g;
	$$self =~ s/\}/&#125;/g;
	$$self;
}
sub detitized
{
	my $self = shift;
	$$self =~ s/&#123;/\{/g;
	$$self =~ s/&#125;/\}/g;
	$$self =~ s/&gt;/>/g;
	$$self =~ s/&lt;/</g;
	$$self =~ s/&quot;/\"/g;
	$$self =~ s/&amp;/&/g;
	$$self;
}
sub linefeed
{ # unify linefeed code
	my $self = shift;
	$$self =~ s/\x0D\x0A/\n/g;
	$$self =~ tr/\x0D\x0A/\n\n/;
	$$self;
}
sub uri_decoded
{ # decode for uri
	my $self = shift;
	$$self =~ s/%([0-9A-Fa-f][0-9A-Fa-f])/pack('H2',$1)/eg;
	$$self =~ s/&#x2F;/\//g;
	$$self;
}
sub uri_encoded
{ # encode for uri
	my $self = shift;
	$$self =~ s/\//&#x2F;/g;
	$$self =~ s/(\W)/'%' . unpack('H2', $1)/eg;
	$$self;
}
sub tag_removed
{
	my $self = shift;
	my $lang = Spica::Language->get;
	my %param = (
		'code'  => $lang->charcode,
		'allow' => undef,
		'fast'  => undef,
		@_
	);
	return( $$self ) if ($$self eq '');
	if ($param{'fast'})
	{
		$$self = &_remove_tag_fast($$self);
	}
	else
	{
		$$self = $lang->convert($$self,DEFAULT_CODE,$param{'code'}) if ($param{'code'} ne DEFAULT_CODE);
		$$self = &_remove_tag_fast($$self,$param{'allow'});
		$$self = $lang->convert($$self,$param{'code'},DEFAULT_CODE) if ($param{'code'} ne DEFAULT_CODE);
	}
	$$self;
}
sub diff
{ # Yuki::DiffText / Copyright (C) 2002 by Hiroshi Yuki.
	# [note] "diff" does not change text itself of the object.
	my $self = shift;
	my $compare = shift;
	return( $$self ) if (!$CAN_DIFF);
	my $diff = undef;
	my @orig = split(/\r?\n/,$compare);
	my @targ = split(/\r?\n/,$$self);
	my (@added,@deleted);
	&Algorithm::Diff::traverse_sequences(
		\@orig, # original array(ref)
		\@targ, # compared array(ref)
		{ # callbacks
			'MATCH'     => \&_matched, # one is the same
			'DISCARD_A' => \&_deleted, # one is only in original
			'DISCARD_B' => \&_added,   # one is only in compared
		}
	);
	&_flush;
	return $diff;
	sub _flush
	{
		$diff .= join('', map { DELETED_LINE . $_ . "\n" } splice(@deleted));
		$diff .= join('', map { ADDED_LINE . $_ . "\n" } splice(@added));
	}
	sub _matched
	{
		&_flush;
		$diff .= SAME_LINE . $orig[$_[0]] . "\n";
	}
	sub _deleted
	{
		push(@deleted,$orig[$_[0]]);
	}
	sub _added
	{
		push(@added,$targ[$_[1]]);
	}
}
# --------------------------------------
# // private functions
sub _remove_tag_fast
{ # it is imperfect but fast to remove html tag from the text
	my $text = shift;
	$text =~ s/<[^>]*>//g; 
	$text;
}
sub _remove_tag_precise
{ # Perl memo <http://www.din.or.jp/~ohzaki/perl.htm> / Copyright (C) 199-2004 OHZAKI Hiroki
	# [note] it is imperfect to remove html tags but it's more precise
	my $text = shift;
	my $tags = shift;
	my ($result,$tmp);
	# templates for regular expression
	my $tag_regex_    = q{[^"'<>]*(?:"[^"]*"[^"'<>]*|'[^']*'[^"'<>]*)*(?:>|(?=<)|$(?!\n))}; #'}
	my $comment_regex = '<!(?:--[^-]*-(?:[^-]+-)*?-(?:[^>-]*(?:-[^>-]+)*?)??)*(?:>|$(?!\n)|--.*$)';
	my $tag_regex     = qq{$comment_regex|<$tag_regex_};
	my $text_regex    = q{[^<]*};
	# removing tags
	while ($text =~ /($text_regex)($tag_regex)?/gso)
	{
		last if ($1 eq '' and $2 eq '');
		$result .= $1;
		$tmp = $2;
		$result .= $tmp if ($tags ne '' and $tmp =~ /^<\/?($tags)(?![0-9A-Za-z])/i);
	}
	$result;
}
sub _clip_text
{
	my ($text,$st,$len) = @_; # input parameters same as substr
	my $lang = Spica::Language->get;
	if ($lang->charcode eq 'utf8')
	{
		return &_clip_text_utf8($text,$st,$len);
	}
	elsif ( $lang->charcode eq 'euc' 
	     or $lang->charcode eq 'sjis' 
	     or $lang->charcode eq 'jis')
	{
		$text = $lang->convert($text,DEFAULT_CODE,$lang->charcode) if ($lang->charcode ne DEFAULT_CODE);
		$text = &_clip_text_euc($text,$st,$len);
		$text = $lang->convert($text,$lang->charcode,DEFAULT_CODE) if ($lang->charcode ne DEFAULT_CODE);
		return($text);
	}
	return substr($text,$st,$len);
}
sub _clip_text_euc
{ # reference : Perl memo http://www.din.or.jp/%7Eohzaki/perl.htm
	my ($text,$st,$len) = @_; # input parameters same as substr
	my $out;
	if ($st > 0)
	{ # checking the start position
		$out = substr($text,0,$st);
		if ($out =~ /\x8F$/)
		{
			$st+=2;
		}
		elsif ($out =~ tr/\x8E\xA1-\xFE// % 2)
		{
			$st++;
		}
	}
	elsif ($st < 0)
	{
		
		$out = substr($text,0,length($text) + $st);
		if ($out =~ /\x8F$/)
		{
			$st--;
		}
		elsif ($out =~ tr/\x8E\xA1-\xFE// % 2)
		{
			$st++;
		}
	}
	$out = substr($text,$st,$len);
	if ($st >= 0)
	{ # adjusting length
		if ($out =~ /\x8F$/)
		{
			$out = substr($text,$st,$len-1);
		}
		elsif ($out =~ tr/\x8E\xA1-\xFE// % 2)
		{
			$out = substr($text,$st,$len+1);
		}
	}
	return($out);
}
sub _clip_text_utf8
{ # reference : http://www.akatsukinishisu.net/itazuragaki/id/round_utf-8
	my ($text,$st,$len) = @_; # input parameters same as substr
	my $out;
	if ($st > 0)
	{ # checking the start position
		$out = substr($text,0,$st);
		UTFST: {
			$_ = $out;
			/[\x00-\x7F]$/, last UTFST;
			/[\xC0-\xFD]$/ and $st--, last UTFST;
			/[\xE0-\xFD][\x80-\xBF]$/ and $st-=2, last UTFST;
			/[\xF0-\xFD][\x80-\xBF]{2}$/ and $st-=3, last UTFST;
		};
		$st = 0 if ($st < 0);
	}
	{ # adjusting length
		$out = substr($text,$st,$len);
		if ($st < 0 and $out !~ /^([\x00-\x7F]|[\xC0-\xFD])/)
		{
			$out =~ s/^[\x80-\xBF]{1,4}//;
		}
		elsif ($out !~ /[\x00-\x7F]$/)
		{
			$out =~ s/[\xC0-\xFD]$//;
			$out =~ s/[\xE0-\xFD][\x80-\xBF]$//;
			$out =~ s/[\xF0-\xFD][\x80-\xBF]{2}$//;
		}
	}
	return($out);
}
1;
__END__

=head1 NAME

Spica::Text - text object

=head1 SYNOPSIS

 use Spica::Text;
 my $text = Spica::Text->new('<a href="http://serennz.sakura.ne.jp/">SerenedipityNZ</a>');
 print $text->entitized; # &lt;a href=&quot; .... &lt;/a&gt;

=head1 DESCRIPTION

=head1 AUTHOR

Takuya Otani <mailto:takuya.otani@gmail.com> http://serennz.sakura.ne.jp/

=head1 LICENSE

Copyright (C) 2005 Takuya Otani@SimpleBoxes / SerendipityNZ Ltd.

=cut
