# Spica::Util::Install
# == Author(s) : Takuya Otani <takuya.otani@gmail.com> ==
# == Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. ==
# == Copyright (C) 2005 Spica Project Kei/Fuco/Takkyun ==

# ======================================
package Spica::Util::Install;
use strict;
use Cwd;
use Spica::Util::Skeleton;
use Spica::File::Object;
use Spica::Driver;
use Spica::Local;
use Spica::Time;
use Spica::Object;
use base qw( Spica::Object );
use vars qw( $VERSION );
$VERSION = '0.03';
# 0.03 [2006/07/23] changed _create_resources to make spica.js
# 0.02 [2006/07/04] got rid of "return undef"
# 0.01 [2006/04/03] changed _get_path to ignore linefeed, _table to handle ";" correctly
# 0.00 [2006/03/31] generated
# --------------------------------------
# // defines / static variables
sub DIR_PERMISSION    (){ 0755 };
sub SCRIPT_PERMISSION (){ 0755 };
sub MODULE_PERMISSION (){ 0644 };
sub TIMEZONE          (){ '+0900' };
my @aDriverType = ('MySQL','Text','SQLite');
# --------------------------------------
# // override base class
sub initialize
{
	my $self = shift;
	my %param = (
		'path'   => $self->_get_path('perl'),
		'time'   => Spica::Time->new,
		'mode'   => undef,
		'step'   => undef,
		'proj'   => undef,
		'data'   => undef,
		'lib'    => undef,
		'script' => undef,
		'zone'   => undef,
		@_
	);
	$self->SUPER::initialize(%param);
	$self->generate_accessor( keys %param );
	$self->_welcome; # display welcome message at first
	return $self->_initialize;
}
# --------------------------------------
# // public functions
sub run
{
	my $self = shift;
	while (<>)
	{
		$_ =~ tr/\x0D\x0A//d;
		next if ($_ =~ /^#/);
		my $result = $self->_parse_input($_);
		last if ($result eq 'exit');
		$self->_prompt;
	}
}
# --------------------------------------
# // private functions - handler
sub _parse_input
{
	my $self = shift;
	my $in = shift;
	return 'exit' if (!$self->mode and $in =~ /^\\q$|^exit$/);
	return $self->_help if (!$self->mode and $in =~ /^\\h$|^help$/);
	return $self->_stat if (!$self->mode and $in =~ /^\\s$|^stat$/);
	return $self->_exec($2) if (!$self->mode and $in =~ /^(\\e|exec)\s+(.+)$/);
	return $self->_create($2) if (!$self->mode and $in =~ /^(\\p|create)\s+(\w+)$/);
	return $self->_define($2) if (!$self->mode and $in =~ /^(\\d|define)\s+(\w+)$/);
	return $self->_project($in) if ($self->mode eq 'proj');
	return $self->_table($in) if ($self->mode eq 'data');
	print 'Illegal command.  Type "\\h" or "help" to display help.',"\n";
}
sub _initialize
{
	my $self = shift;
	if (!-e $self->lib)
	{ # create directory
		Spica::File::Dir->new('path'=>$self->lib)->touch(DIR_PERMISSION);
	}
	unshift(@INC,$self->lib) if (-e $self->lib);
	my $check = $self->lib . 'config.pl';
	if (-e $check)
	{ # config.pl already exists, it means project is already created
		require $check;
		my $conf = Spica::Config->new;
		$self->proj(Spica::Driver->new($conf->value('driver'),%{$conf->value('db_param')}));
	}
	$self->_prompt;
	$self;
}
# --------------------------------------
# // private functions - project
sub _create
{
	my $self = shift;
	my $name = shift;
	if (!$self->proj)
	{
		$Spica::PRODUCT = $name;
		$self->mode('proj');
		$self->step(1);
		$self->proj({}); # initialize proj
		$self->_project_prompt('creating project ... ' . $name);
	}
	else
	{
		print 'project "',$Spica::PRODUCT,'" is already created.',"\n";
		$self->mode(undef);
		$self->step(undef);
	}
}
sub _project
{
	my $self = shift;
	my $in = shift;
	my $err = undef;
	if ($in eq '\\q')
	{
		$err = 'quit to create a project';
		$self->mode(undef);
		$self->step(undef);
		$self->proj(undef);
	}
	elsif ($in eq '\\h')
	{
		$err = 'configure the project following prompt';
	}
	elsif ($self->step == 1)
	{
		$err = 'wrong number' if ($in eq '' or $in =~ /\D/ or !defined($aDriverType[$in]));
		if (!$err)
		{
			my $module = $self->_module_name($aDriverType[$in]);
			my $class = Spica::Driver->module_name($module);
			if (&Spica::Object::load_module($class))
			{
				my $num = 0;
				$self->proj->{'driver'} = $module;
				$self->proj->{'db_param'} = {$class->driver_param};
				$self->proj->{'_check'} = 0;
				$self->proj->{'_list'} = [grep( ++$num % 2 , $class->driver_param )];
				$self->step(2);
			}
			else
			{ # failed to load
				$err = $@;
			}
		}
	}
	elsif ($self->step == 2)
	{
		$err = 'wrong value' if ($in ne '' and $in =~ m/[^a-zA-Z0-9_\-\\\.\/]/);
		if (!$err)
		{
			my $key = shift(@{$self->proj->{'_list'}});
			$self->proj->{'db_param'}->{$key} = $in if ($in ne '');
			$err = $self->_check_project if ($#{$self->proj->{'_list'}} == -1);
		}
	}
	$self->_project_prompt($err);
}
sub _project_prompt
{
	my $self = shift;
	my $msg = shift;
	print $msg,"\n" if ($msg);
	if ($self->step == 1)
	{
		print '* select driver ';
		for (my $i=0;$i<@aDriverType;$i++)
		{
			printf('[%d]%s ',$i,$aDriverType[$i]);
		}
		print '?',"\n";
	}
	elsif ($self->step == 2)
	{
		my $key = $self->proj->{'_list'}->[0];
		print '* input ',$key,' [default => ',$self->proj->{'db_param'}->{$key},'] ?',"\n";
	}
}
sub _check_project
{
	my $self = shift;
	my $driver = undef;
	eval
	{
		$driver = Spica::Driver->new(
			$self->proj->{'driver'},
			%{$self->proj->{'db_param'}},
		);
	};
	if ($driver)
	{
		Spica::File::Dir->new('path'=>$self->lib . 'Spica')->touch(DIR_PERMISSION);
		Spica::File::Dir->new('path'=>$self->lib . 'Spica/App')->touch(DIR_PERMISSION);
		Spica::File::Dir->new('path'=>$self->lib . 'Spica/Data')->touch(DIR_PERMISSION);
		Spica::File::Dir->new('path'=>$self->lib . 'resource')->touch(DIR_PERMISSION);
		$self->_create_config($driver);
		$self->_create_script;
		$self->_create_module;
		$self->_create_resources;
		$self->proj($driver);
		$self->step(undef);
		$self->mode(undef);
		return 'project "' . $Spica::PRODUCT . '" is created!';
	}
	return $@;
}
sub _create_config
{
	my $self = shift;
	my $driver = shift;
	my $file = Spica::File->new('path'=>$self->lib,'name'=>'config.pl');
	my $date = $self->time->format('form'=>'%Year%/%Mon%/%Day%','zone'=>$self->_zone);
	my %check = $driver->driver_param;
	my $base = Spica::Util::Skeleton->skeleton('config_base');
	my $conf = '';
	$conf .= "\t\t'driver'       => '" . $self->proj->{'driver'} . "',\n";
	$conf .= "\t\t'db_param'     => {";
	foreach my $key ( keys %{$self->proj->{'db_param'}} )
	{
		next if ($self->proj->{'db_param'}->{$key} eq $check{$key});
		$conf .= ($self->proj->{'db_param'}->{$key} =~ /^\-?\d+$/)
			? "'$key'=>" . $self->proj->{'db_param'}->{$key} . ','
			: "'$key'=>'" . $self->proj->{'db_param'}->{$key} . "',";
	}
	$conf .= '},' . "\n";
	$conf .= "\t\t'resource_dir' => 'lib/resource/',\n";
	$conf .= "\t\t'site_css'     => 'lib/resource/style.css',\n";
	$conf .= "\t\t'server_path'  => 'http://127.0.0.1/',";
	$base =~ s/__PROJECT_NAME__/$Spica::PRODUCT/g;
	$base =~ s/__GENERATED_DATE__/$date/g;
	$base =~ s/__CONFIGURATION__/$conf/g;
	$file->touch(MODULE_PERMISSION);
	$file->set_data($base);
}
sub _create_script
{
	my $self = shift;
	my $file = Spica::File->new('path'=>'./','name'=>$self->script);
	my $base = Spica::Util::Skeleton->skeleton('script_base');
	my $path = $self->path;
	$base =~ s/__PROJECT_NAME__/$Spica::PRODUCT/g;
	$base =~ s/__PERL_PATH__/$path/g;
	$file->touch(SCRIPT_PERMISSION);
	$file->set_data($base);
}
sub _create_module
{
	my $self = shift;
	my $file = Spica::File->new('path'=>$self->lib . 'Spica/App','name'=>'Main.pm');
	my $base = Spica::Util::Skeleton->skeleton('app_base');
	my $date = $self->time->format('form'=>'%Year%/%Mon%/%Day%','zone'=>$self->_zone);
	$base =~ s/__PROJECT_NAME__/$Spica::PRODUCT/g;
	$base =~ s/__GENERATED_DATE__/$date/g;
	$file->touch(MODULE_PERMISSION);
	$file->set_data($base);
}
sub _create_resources
{
	my $self = shift;
	my %resource = (
		'main.html' => 'template_base',
		'spica.png' => 'logo_base',
		'style.css' => 'style_base',
		'spica.js'  => 'javascript_base',
		'ja.txt'    => 'string_base',
	);
	foreach my $key ( keys %resource )
	{
		my $file = Spica::File->new('path'=>$self->lib . 'resource','name'=>$key);
		$file->touch(MODULE_PERMISSION);
		$file->set_data(Spica::Util::Skeleton->skeleton($resource{$key}));
	}
}
# --------------------------------------
# // private functions - data object
sub _define
{
	my $self = shift;
	my $name = shift;
	$name = $self->_module_name($name);
	my $file = $self->lib . 'Spica/Data/' . $name . '.pm';
	if ($self->proj and !-e $file)
	{
		$self->mode('data');
		$self->step($name);
		$self->data({'list'=>['id'],'info'=>{'id'=>'uni_id'},'max'=>2}); # initialize data
		$self->_table_prompt('creating table ... ' . $name);
	}
	else
	{
		print 'need to create a project first',"\n" if (!$self->proj);
		print 'data module is already created',"\n" if (-e $file);
		$self->mode(undef);
		$self->step(undef);
	}
}
sub _table
{
	my $self = shift;
	my $in = shift;
	my $err = undef;
	my $num = 0;
	my @types = grep( ++$num % 2, Spica::Driver->table_param);
	if ($in eq '\\q')
	{
		$err = 'quit to create a table';
		$self->mode(undef);
		$self->step(undef);
	}
	elsif ($in eq '\\h')
	{
		$err  = 'input element name and type as follow:' . "\n";
		$err .= '  ELEMENT_NAME ELEMENT_TYPE' . "\n";
		$err .= 'input ";" to finish' . "\n\n";
		$err .= 'element type can be chosen from the following options:'."\n";
		$err .= '  ' . join(',',@types);
	}
	elsif ($in eq ';')
	{
		$err = $self->_create_table;
	}
	else
	{
		$in =~ s/,/ /g;
		$in =~ s/=>/ /g;
		$in =~ s/^\s*//;
		$in =~ s/\s*$//;
		if ($in =~ /(\w+)\s+(.+)/)
		{
			my @buf = @{$self->data->{'list'}};
			my ($name,$type) = ($1,$2);
			$type =~ s/;//g;
			$err = $name . ' is already defined' if ( grep(/^\Q$name\E$/,@buf) );
			if (!$err and $type =~ /(\w+)/ and grep(/^$1$/,@types))
			{
				push(@{$self->data->{'list'}},$name);
				$self->data->{'info'}->{$name} = $type;
				$self->data->{'max'} = length($name) if (length($name) > $self->data->{'max'});
				$err = $self->_create_table if ($in =~ /;/);
			}
			elsif (!$err)
			{
				$err = 'wrong type : ' . $type;
			}
		}
		else
		{
			$err = 'wrong value : ' . $in;
		}
	}
	$self->_table_prompt($err);
}
sub _create_table
{
	my $self = shift;
	my $name = $self->step;
	my $path = $self->lib . 'Spica/Data/' . $name . '.pm';
	my $file = Spica::File->new('fullpath'=>$path);
	my $date = $self->time->format('form'=>'%Year%/%Mon%/%Day%','zone'=>$self->_zone);
	my $base = Spica::Util::Skeleton->skeleton('data_base');
	my $data = '';
	print 'creating [',$name,'] ... ';
	foreach my $key ( @{$self->data->{'list'}} )
	{
		$data .= "\t\t'" . $key . "' " . (' ' x ($self->data->{'max'} - length($key)));
		$data .= "=> '" . $self->data->{'info'}->{$key} . "',\n";
	}
	$data =~ s/\n$//;
	$base =~ s/__DATA_CLASS__/$name/g;
	$base =~ s/__GENERATED_DATE__/$date/g;
	$base =~ s/__DATA_STRUCTURE__/$data/g;
	$file->touch(MODULE_PERMISSION);
	$file->set_data($base);
	print ' done',"\n",'creating table in database ... ';
	require $path;
	$self->proj->create_table('Spica::Data::' . $name);
	$self->mode(undef);
	$self->step(undef);
	print 'done',"\n";
	return '"Spica::Data::' . $name . '" is created!';
}
sub _table_prompt
{
	my $self = shift;
	my $msg = shift;
	print $msg,"\n" if ($msg);
}
# --------------------------------------
# // private functions - path
sub _get_path
{ # taken from server analizer / Copyright (C) futomi 2003 - 2005
	my $self = shift;
	my $target = shift;
	my @path = ();
	return if (!$target);
	return $self->_get_perlpath_onwin($target) if ($^O =~ /MSWin32/i);
	foreach my $whereis ('whereis', '/usr/bin/whereis', '/usr/ucb/whereis')
	{
		my $res = `$whereis $target`;
		next if (!$res);
		my @list = split(/\s/,$res);
		foreach my $check (@list)
		{
			push(@path,$check) if ($check =~ m!/${target}$!);
		}
		last;
	}
	if (!@path)
	{
		my $res = `which $target`;
		push(@path,$res) if ($res =~ m!${target}$!);
	}
	@path = map { s/\n$//; $_ } @path;
	return $path[0]; # return just top one
}
sub _get_perlpath_onwin
{ # taken from server analizer / Copyright (C) futomi 2003 - 2005
	my $self = shift;
	my $target = shift;
	return if ($target ne 'perl');
	my $path = $^X;
	$path =~ s!\\!/!g;
	$path =~ s!\.exe$!!;
	$path;
}
sub _get_tables
{
	my $self = shift;
	my $dir = Spica::File::Dir->new('path'=>$self->lib . 'Spica/Data');
	return if (!$dir->exists);
	my @list = $dir->get_files;
	return map { $_->name_without_ext } @list;
}
# --------------------------------------
# // private functions - display
sub _module_name
{
	my $self = shift;
	my $name = shift;
	$name =~ s/^(\w)(\w+)$/uc($1).lc($2)/e;
	$name;
}
sub _zone
{
	my $self = shift;
	return TIMEZONE if ($self->zone eq '');
	return Spica::Local->timezone($self->time,$self->zone);
}
sub _prompt
{
	my $self = shift;
	my $prompt = ($self->mode) 
		? '    :> ' 
		: 'spica> ';
	print $prompt;
}
sub _output
{
	my $self = shift;
	my $data = shift;
	$data =~ s/^\s+//;
	$data =~ s/\n\s+/\n/g;
	$data =~ s/^# //;
	$data =~ s/\n# /\n/g;
	$data;
}
sub _exec
{
	my $self = shift;
	my $command = shift;
	my $result = `$command`;
	print $result;
	print "\n" if ($result and $result !~ /\n$/);
	print 'done : ' . $command . "\n" if ($result eq '');
}
sub _stat
{
	my $self = shift;
	print 'current directory : ',cwd(),"\n";
	print 'current project : ',($self->proj) ? $Spica::PRODUCT : '[undefined]',"\n";
	if ($self->proj)
	{
		my @tables = $self->_get_tables;
		print 'database driver : ',Spica::Config->get->value('driver'),"\n";
		print 'data tables : ',"\n -> ",join("\n -> ",@tables),"\n" if (@tables);
	}
}
sub _help
{
	my $self = shift;
	print $self->_output(<<"__HELP_MESSAGE__");
	# \\q or exit : exit helper
	# \\h or help : display help
	# \\s or stat : display status
	# \\e or exec COMMAND : execute external command
	# \\p or create PROJECT_NAME : make project. PROJECT_NAME must be a-zA-Z0-9_.
	# \\d or define TABLE_NAME : make data. TABLE_NAME must be a-zA-Z0-9_.
__HELP_MESSAGE__
}
sub _welcome
{
	my $self = shift;
	print $self->_output(<<"__WELCOME_MESSAGE__");
	# 
	# ==/ Spica - Simple Primary Interface and Core Application /==
	# ----- Copyright (C) 2005 SimpleBoxes/SerendipityNZ Ltd. -----
	# ----- Copyright (C) 2005 Spica Project Kei/Fuco/Takkyun -----
	# ======================================= Spica Version : $Spica::LIBRARY
	# 
__WELCOME_MESSAGE__
}
1;
__END__
